module FileSystemLayerBase


open FlashFileSystemCore
/********************************************************************
 * authors: 
 *   Miguel Ferreira    <miguel@di.uminho.pt>
 *   Samuel Silva       <silva.samuel@alumni.uminho.pt>
 * status: experimental (unstable, experimental, stable)
 * description: 
 *     
 */

/*******************************************************************/

open RelCalc

/********************************************************************
 * Type Translation
 */
/***********************************************/

-- captures the entire file system
sig System {
  table:     OpenFilesTable, -- where open files are registered
  fileStore: FileStore       -- where files are stored
}

pred SystemInvariantVDM[sys: System] {
  OpenFilesTableInvariantVDM[sys.table] and
  FileStoreInvariantVDM[sys.fileStore]  and
  SystemInvariant[sys]
}

pred SystemInvariant[s: System]{
  RelCalc/rng[s.table.map].path in RelCalc/dom[s.fileStore.map]
} 
  -- all open files must exist 
  -- | in the file store


-- captures the file storage sub system
sig FileStore {
  map: Path -> File
}

pred FileStoreInvariantVDM[fs: FileStore]{
  RelCalc/Simple[fs.map, File]          and
  RelCalc/Injective[fs.map, Path]       and   -- in VDM the mk_ contruct is injective
  PathInvariantVDM[RelCalc/dom[fs.map]] and 
  FileInvariantVDM[RelCalc/rng[fs.map]] and
  FileStoreInvariant[fs]
}

pred FileStoreInvariant[t: FileStore]{
  (t.map).(File->Directory) in (dirName).(t.map).attributes.fileType
}


-- captures the full paths the point to files
abstract sig Path {
  dirName: Path
}

-- captures a sequence of file names
sig FileNames extends Path {} 

one sig Root extends Path {}

pred PathInvariantVDM[p: Path]{
  RelCalc/Function[dirName,Path,Path] and
  PathInvariant[p]
}

pred PathInvariant[p: Path]{
  dirNameProperties
}

-- defines dirName properties
pred dirNameProperties {
  RelCalc/Reflexive[(RelCalc/id[Root]).(dirName),Root]
  RelCalc/Acyclic[(RelCalc/id[FileNames]).(dirName),FileNames]
}


-- captures a file (in its wider sense)
sig File {
  attributes: Attributes,  -- file's attributes
  contents:   FileContents -- file's contents
}

pred FileInvariantVDM[f: File]{
  RelCalc/Injective[attributes,File]
  RelCalc/Injective[contents,File]
  FileInvariant[f]
}

pred FileInvariant[f: File] {
  all file: f {
    (file.attributes.fileType in Directory and
     file.contents in Nothing)    -- directories can't have contents
    or
    (file.attributes.fileType in RegularFile and
     file.contents in Contents)   -- regular files do have contents 
  }
}


-- captures files attributes
-- | (for now abstracts the FS_FileDirInfo [15.2.6])
sig Attributes {
  fileType: FileType       -- files type
}

-- captures the diferent kinds of file types
-- | regular files (the common files) are a kind of file
-- | directories are also a kind of file
abstract sig FileType {}

one sig RegularFile extends FileType {}

one sig Directory extends FileType {}


-- captures the contents of a file
abstract sig FileContents {}

sig Contents extends FileContents {}    -- some contents

one sig Nothing extends FileContents {} -- no contents


-- captures the open files management sub system
sig OpenFilesTable {
  map: FS_FileHandle -> FS_OpenFileInfo,
}

pred OpenFilesTableInvariantVDM[b: OpenFilesTable]{
  RelCalc/Simple[b.map,FS_OpenFileInfo]           and
  FS_OpenFileInfoInvariantVDM[RelCalc/rng[b.map]]
}


-- captures a reference for an open file
-- page 260
sig FS_FileHandle {} -- abstracts the pointer

-- captures the information on a open file
-- page 268 -- 15.2.7 FS_OpenFileInfo
sig FS_OpenFileInfo {
  accessMode:  FS_AccessMode,     -- The mode in which the file was opened
  modified:    Bool,              -- Indicates if the file was changed
  sizeChanged: Bool,              -- True if:
                                  -- | + file size is changed
                                  -- | + (new size entry in the dynamic info object allocated)
  path:        Path
}

pred FS_OpenFileInfoInvariantVDM[ofi: FS_OpenFileInfo]{
  RelCalc/Injective[path,FS_OpenFileInfo]        and
  RelCalc/Injective[accessMode,FS_OpenFileInfo]  and
  PathInvariantVDM[ofi.path]
}


-- auxiliar predicates --

-- checks if file is a directory
pred isDirectory[f: File] {
  f.attributes.fileType in Directory
}

-- checks if file is a regular file
pred isRegularFile[f: File] {
  f.attributes.fileType in RegularFile
}

-- checks if file has contents
pred hasContents[f: File] {
  f.contents not in Nothing
}

-- checks if a file (pointed by path) has sub files
pred hasSubFiles[fs: FileStore, path: Path] {
  path in RelCalc/dom[fs.map].dirName
}

-- checks if path is the Root
pred isRoot[path: Path] {
  path in Root
}

-- checks if path belongs to the fileStore
pred isElemFileStore[path: Path, fs: FileStore] {
  path in RelCalc/dom[fs.map]
}

-- checks if the file pointed by path is open
pred isElemTable[p: Path, table: OpenFilesTable] {
  p in RelCalc/rng[table.map].path
}

-- check if handle belongs to the table
pred isElemTable[handle: FS_FileHandle, table: OpenFilesTable] {
  handle in RelCalc/dom[table.map]
}

-- check is parent is path's dirName
pred isDirName[parent, path: Path] {
  parent = path.dirName
}

-- given a file type returns the appropriate contets
pred emptyFileContents[ft: FileType, fc: FileContents] {
  ft = Directory => fc = Nothing else fc = Contents
}

-- returns a new and distinct handle 
pred newFileHandle[handles: set FS_FileHandle, handle: FS_FileHandle] {
  handle not in handles
}


--- Show Predicates -------

pred ShowFileStore[fs: FileStore] {
  some fs.map                                        and
  Path = RelCalc/dom[fs.map]                         and
  File = RelCalc/rng[fs.map]                         and
  Attributes = RelCalc/rng[fs.map].attributes        and
  FileType = RelCalc/rng[fs.map].attributes.fileType and
  PathInvariantVDM[Path]                             and        
  FileStoreInvariantVDM[fs]
}

run ShowFileStore for 6 but 1 FileStore,
                            0 System,
                            0 OpenFilesTable,
                            0 FS_FileHandle,
                            0 FS_OpenFileInfo

pred ShowOpenFilesTable[table: OpenFilesTable] {
  some table.map                           and
  FS_FileHandle = RelCalc/dom[table.map]   and
  FS_OpenFileInfo = RelCalc/rng[table.map] and
  OpenFilesTableInvariantVDM[table]
}

run ShowOpenFilesTable for 3 but 1 OpenFilesTable,
                                 0 System,
                                 0 FileStore,
                                 0 File

pred ShowSystem[sys: System] {
  ShowFileStore[sys.fileStore]  and
  ShowOpenFilesTable[sys.table] and
  OpenFilesTable = sys.table    and
  FileStore = sys.fileStore     and
  SystemInvariantVDM[sys]
} 

run ShowSystem for 3 but 1 System

