module FlashFileSystemCore

open util/integer
open util/boolean



/********************************************************************
 * authors: 
 *   Miguel Ferreira    <miguel@di.uminho.pt>
 *   Samuel Silva       <silva.samuel@alumni.uminho.pt>
 * status: experimental (unstable, experimental, stable)
 * description: 
 *     
 */

-- captures the status of an operation
sig FFS_Status {}

one sig FFS_StatusSuccess         extends FFS_Status {} 
one sig FFS_ErrorSystem           extends FFS_Status {}
one sig FS_ErrorFileNotFound      extends FFS_Status {}
one sig FS_ErrorFileStillOpen     extends FFS_Status {}
one sig FS_ErrorInvalidPath       extends FFS_Status {}
one sig FS_ErrorFileNotOpen       extends FFS_Status {}
one sig FS_ErrorInvalidHandle     extends FFS_Status {}
one sig FS_ErrorFileAlreadyExists extends FFS_Status {}
one sig FS_ErrorDirectoryNonEmpty extends FFS_Status {}
one sig FFS_StatusUnknown         extends FFS_Status {}

-- captures the modes that can be used to open files
-- page 261 -- 15.2.9 FS_OpenMode
-- * FS_OpenMode defines all allowable modes in which a file can be opened, when using the 
-- | FS_OpenFileDir() API function.  
sig FS_OpenMode {}

one sig FS_CreateNew            extends FS_OpenMode {}
one sig FS_CreateAlways         extends FS_OpenMode {}
one sig FS_OpenRead             extends FS_OpenMode {}
one sig FS_OpenWrite            extends FS_OpenMode {}
one sig FS_OpenAlways           extends FS_OpenMode {}
one sig FS_OpenWriteOnly        extends FS_OpenMode {}
one sig FS_CreateAlwaysReadOnly extends FS_OpenMode {}
one sig FS_CreateNewReadOnly    extends FS_OpenMode {}


-- captures the modes that can be used to access files
-- page 262 -- 15.2.10 FS_AccessMode
-- * FS_AccessMode defines how a file is opened.
-- | This is used for validation of operations on the open file.
sig FS_AccessMode {}

one sig  FS_AccessReadOnly  extends FS_AccessMode {}
one sig  FS_AccessWriteOnly extends FS_AccessMode {}
one sig  FS_AccessReadWrite extends FS_AccessMode {}


-- captures the information on the underlying hardware partitions information
-- page 255 -- 15.2.4 FFS_VolumeInfo
-- * The FFS_VolumeInfo structure stores space usage 
-- | information about the file system
-- * fragments <=> ONFI block
sig FFS_VolumeInfo {
  totalFragmentsts        : Int, -- This equals the total fragments in the filesystem
  totalAvailableFragments : Int, -- TotalAvailableFragments is the size of the fragment
                                 -- | granularity that was defined by the user
  totalInvalidFragments   : Int, -- This denotes the number of dirty fragments in the volume.
                                 -- | When a file is deleted, the data/meta
                                 -- | data of that file is no longer valid and denoted as “dirty.”
  fragmentGranularity     : Int  -- This is the value, in bytes, of the fragment granularity.
}

-- captures the configurable options of a volume
-- page 254 -- 15.2.2 FFS_Format
-- * The FFS_Format structure is used to provide user-configurable
-- | format options in the formation of a logical volume in flash.
sig FFS_Format {
  fragmentGranularity      : Int, -- Defines the size, in bytes, of a logical unit of allocation 
                                  -- | as the smallest granularity of data storage space that the 
                                  -- | file system can allocate
  fragmentOverheadPerBlock : Int  -- Defines the amount of overhead per flash erase block within 
                                  -- | the volume
}

pred inv_FFS_Format[ffs_format: FFS_Format]{ 
  fragmentGranularity_AllowedSizes[ffs_format.fragmentGranularity]
}

-- captures the configurable options of the file system
-- page 254 -- 15.2.3 FFS_Options
-- * The FFS_Options structure provides user-configurable file system 
-- | options upon system initialization
sig FFS_Options {
  maxOpenFiles           : Int,  -- The maximum number of concurrently open files that the file 
                                 -- | system can manage at a time.
                                 -- | POSIX specs recommends a value of 20
  maxPathLength          : Int,  -- The maximum length of a file / directory path, in (bytes) chars
                                 -- | POSIX specs recommends a value of 1024
  timeStampEnabled       : Bool, -- When set to TRUE, a file or directory's time and date stamp is  
                                 -- | updated every time it is accessed, or when its contents are updated
  userReclaimedThreshold : Int   -- Threshold, configured as the percent dirty, that will be 
                                 -- | used by the external user reclaim API
}

pred inv_FFS_Options[ffs_options: FFS_Options] {
  maxPathLength_AllowedValues[ffs_options.maxPathLength]
}

-- maps open modes to access modes
pred fs_open2access_mode_map[omode: FS_OpenMode, amode: FS_AccessMode] {
  (omode = FS_CreateNew            and amode = FS_AccessReadWrite) or
  (omode = FS_CreateAlways         and amode = FS_AccessReadWrite) or
  (omode = FS_OpenRead             and amode = FS_AccessReadOnly)  or
  (omode = FS_OpenWrite            and amode = FS_AccessReadWrite) or
  (omode = FS_OpenAlways           and amode = FS_AccessReadWrite) or
  (omode = FS_OpenWriteOnly        and amode = FS_AccessWriteOnly) or
  (omode = FS_CreateAlwaysReadOnly and amode = FS_AccessReadOnly)  or
  (omode = FS_CreateNewReadOnly    and amode = FS_AccessReadOnly)
}

-- defines the valid fragment granularity values
pred fragmentGranularity_AllowedSizes[size: Int]{
  size = 1024 or
  size = 2048 or
  size = 4096 or
  size = 8192
}

-- defines the max path length values
pred maxPathLength_AllowedValues[length : Int]{
  length   >= 8          and
  length   <= 1024       and
  pre_multiple[length,4] and
  length.multiple[4]
}

-- Allowable range is 8 to 1024, in 
-- | multiples of 4.  Values that do not divide evenly by 4 are rounded up to the next 
-- | nearest multiple of 4.

-- definition of multiple nats
-- x is multiple of y
pred multiple[x,y : Int]  {
  0 = int[x].rem[int[y]]
}

pred pre_multiple[x,y: Int] {
  int[y] not = 0
}


pred Show[] {
  all ffs_format:  FFS_Format  { inv_FFS_Format[ffs_format] }
  all ffs_options: FFS_Options { inv_FFS_Options[ffs_options] }
}

--run Show for 15 int
